### The simple RL example from Levin et al, 2001.
### Michael Gasser, gasser@cs.indiana.edu
### Indiana University
### 4/2012

from qlearn import *

# Months and days
MONTHS = ['jun', 'jul', 'aug']
DAYS = range(1, 32)
# June has 31 days
DATES = [('jun', d) for d in DAYS] + \
        [('jul', d) for d in DAYS] + \
        [('aug', d) for d in DAYS]
STATES = ['start', 'end'] + \
         [(m, None) for m in MONTHS] + \
         [(None, d) for d in DAYS] + \
         DATES

# The action set
ACTIONS = ['ask_month', 'ask_day', 'ask_date', 'close']

# Probabilities of miscomprehension
MONTH_ERROR = .05
DAY_ERROR = .1
DATE_ERROR = .15

# Weights for different parts of the cost function
W_LENGTH = 1.0
W_ERROR = 1.5
W_GOALDIST = 1.5

def reinforcement(errors, goaldist):
    """Reinforcement for # errors and distance from goal."""
    return -(W_LENGTH + W_ERROR * errors + W_GOALDIST * goaldist)

def get_goaldist(state):
    """Distance from state in number of components not filled."""
    if month_state(state) or day_state(state):
        return 1
    elif date_state(state):
        return 0
    else:
        return 2

def get_month_errors(state, month):
    """Errors in the month part of state."""
    if state[0] != month:
        return 1
    else:
        return 0

def get_day_errors(state, day):
    """Errors in the day part of state."""
    if state[1] != day:
        return 1
    else:
        return 0

def get_date_errors(state, month, day):
    """Errors in both parts of state."""
    errors = 0
    if state[0] != month:
        errors += 1
    if state[1] != day:
        errors += 1
    return errors

def month_state(state):
    """Is this a month-only state?"""
    return isinstance(state, tuple) and state[0] and not state[1]
def day_state(state):
    """Is this a day-only state?"""
    return isinstance(state, tuple) and state[1] and not state[0]
def date_state(state):
    """Is this a state with both month and day?"""
    return isinstance(state, tuple) and state[0] and state[1]

def prob_choice(choices, correct, prob):
    """The correct choice or a random selection from choices, depending on prob."""
    if random.random() < prob:
        return random.choice(choices)
    else:
        return correct

def gen_date():
    """Generate a random date."""
    return random.choice(DATES)

def next_state_reinf(state, action, date):
    """The next state and reinforcement, given state, action, and the goal date."""
    month, day = date
    if state == 'end':
        return False, False
    if action == 'close':
        return 'end', reinforcement(0, get_goaldist(state))
    if action == 'ask_month':
        state = (prob_choice(MONTHS, month, MONTH_ERROR), state[1] if isinstance(state, tuple) else None)
        errors = get_month_errors(state, month)
        return state, reinforcement(errors, 0)
    if action == 'ask_day':
        state = (state[0] if isinstance(state, tuple) else None, prob_choice(DAYS, day, DAY_ERROR))
        errors = get_day_errors(state, day)
        return state, reinforcement(errors, 0)
    if action == 'ask_date':
        state = (prob_choice(MONTHS, month, MONTH_ERROR), prob_choice(DAYS, day, DATE_ERROR))
        errors = get_date_errors(state, month, day)
        return state, reinforcement(errors, 0)

levin_learner = Learner(STATES, ACTIONS, next_state_reinf, 'start', gen_date)
